/*****************************************************************************
 * encoder.c: video encoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008-2009 M2X BV
 * $Id: f3b753242fa86fe9264e493a6ee4be96742cda09 $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Rafaël Carré <rcarre@m2x.nl>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_vout.h>
#include <vlc_aout.h>
#include <vlc_sout.h>
#include <vlc_codec.h>

#include "davinci.h"

#include <ti/sdo/ce/video/videnc.h>

#include <assert.h>

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static block_t *EncodeVideo( encoder_t *, picture_t * );

/*****************************************************************************
 * Video encoder
 *****************************************************************************/
struct encoder_sys_t
{
    char *psz_ti_engine;

    Engine_Handle e;
    VIDENC_Handle c;

    XDM_BufDesc in;
    XDM_BufDesc out;
};

/*****************************************************************************
 *
 *****************************************************************************/
int OpenVideoEncoder( vlc_object_t *p_this )
{
    encoder_t *p_enc = (encoder_t *)p_this;
    encoder_sys_t *p_sys;
    Engine_Error err;
    VIDENC_Params params;
    int i_cat;
    const char *psz_codec;
    const char *psz_namecodec;

    if( !GetDavinciEncoder( p_enc->fmt_out.i_codec, &i_cat,
                            &psz_codec, &psz_namecodec ) )
    {
        msg_Dbg( p_enc, "Unsupported codec : %4.4s",(char*)&p_enc->fmt_out.i_codec );
        return VLC_EGENERIC;
    }

    msg_Dbg( p_enc, "found davinci encoder %s for %s (codec=%4.4s, cat=%s)",
                psz_codec, (char*)&p_enc->fmt_out.i_codec, psz_namecodec,
                (i_cat == VIDEO_ES) ? "video" :
                    (i_cat == AUDIO_ES) ? "audio" : "other" );

    /* Allocate our private structure */
    p_enc->p_sys = (encoder_sys_t *)calloc( 1, sizeof( encoder_sys_t ) );
    if( !p_enc->p_sys )
        return VLC_ENOMEM;
    p_sys = p_enc->p_sys;

    p_sys->psz_ti_engine = var_CreateGetString( p_enc, DEC_CFG_PREFIX "engine" );
    if( !p_sys->psz_ti_engine )
    {
        free( p_sys );
        return VLC_ENOMEM;
    }

    /* FIXME: number of frames between 2 keyframes */
    p_enc->i_iframes = var_CreateGetInteger( p_enc, ENC_CFG_PREFIX "keyint" );
    p_enc->i_tolerance = var_CreateGetInteger( p_enc, ENC_CFG_PREFIX "tolerance" );
    p_enc->i_tolerance *= 1024; /* bits per second */

    /* Initialize the codec engine */
    CERuntime_init();

    /* Create an engine handle */
    p_sys->e = Engine_open( p_sys->psz_ti_engine, NULL /*&Engine_ATTRS*/, &err );
    if( err != Engine_EOK )
    {
        msg_Err( p_enc, "Error while opening engine `%s': %s",
                 p_sys->psz_ti_engine, ppsz_engine_error[err] );
        goto error;
    }

    davinci_PrintAvailableAlgorithms( p_this, p_sys->psz_ti_engine );

    /* Configure encoder */
    params.size = sizeof( params );
    switch( var_CreateGetInteger( p_enc, ENC_CFG_PREFIX "quality" ) )
    {
        case 1: params.encodingPreset = XDM_HIGH_QUALITY; break;
        case 2: params.encodingPreset = XDM_HIGH_SPEED; break;
        case 3: params.encodingPreset = XDM_USER_DEFINED; break;
        default:
            params.encodingPreset = XDM_DEFAULT;
            break;
    }
    switch( var_CreateGetInteger( p_enc, ENC_CFG_PREFIX "rc" ) )
    {
        case 1: params.rateControlPreset = IVIDEO_LOW_DELAY; break;
        case 2: params.rateControlPreset = IVIDEO_STORAGE; break;
        case 3: params.rateControlPreset = IVIDEO_TWOPASS; break;
        case 4: params.rateControlPreset = IVIDEO_NONE; break;
        case 5: params.rateControlPreset = IVIDEO_USER_DEFINED; break;
        default:
            params.rateControlPreset = IVIDEO_RATECONTROLPRESET_DEFAULT;
            break;
    }
    params.maxHeight = p_enc->fmt_in.video.i_height;
    params.maxWidth = p_enc->fmt_in.video.i_width;
    params.maxFrameRate = p_enc->fmt_out.video.i_frame_rate_base ?
                          p_enc->fmt_out.video.i_frame_rate * 1000 /
                          p_enc->fmt_out.video.i_frame_rate_base :
                          50000; /* Frames per 1000 seconds  */
    params.maxBitRate = p_enc->fmt_out.i_bitrate
                      + p_enc->i_tolerance;

    params.dataEndianness = XDM_BYTE;
    params.maxInterFrameInterval = 0; /* NOTE: should be null */
    params.inputChromaFormat = VlcChromaToXdm( p_enc->fmt_in.i_codec );

    msg_Info( p_enc, "max %dx%d at %.3f fps (max bitrate %d kBps, I-Frame interval %d)\n",
        (int)params.maxWidth, (int)params.maxHeight,
        ((float)params.maxFrameRate)/1000.,
        ((int)params.maxBitRate) / (8*1024),
        (int)p_enc->i_iframes );

    if( params.inputChromaFormat == XDM_CHROMA_NA )
    {
        msg_Warn( p_enc, "Unsupported input chroma (%4.4s), forcing I420.",
                 (const char *)&p_enc->fmt_in.i_codec );
        params.inputChromaFormat = VlcChromaToXdm( VLC_FOURCC('I','4','2','0' ) );
    }

    /* NOTE: we don't know if it's progressive or interlaced
     * until we get the first picture (p_pic->b_progressive).
     * Thus let a knowledgeable user override this setting. */
    if( var_CreateGetBool( p_enc, ENC_CFG_PREFIX "interlace" ) )
        params.inputContentType = IVIDEO_INTERLACED;
    else
        params.inputContentType = IVIDEO_PROGRESSIVE;

    /* Create encoder handle */
    p_sys->c = VIDENC_create( p_sys->e, (String)psz_codec, &params );
    if( !p_sys->c )
    {
        msg_Err( p_enc, "Failed to create video encoder (%s) for %s",
                 psz_codec, psz_namecodec );
        goto error;
    }

    /* Initialize random stuff */
    p_enc->pf_encode_video = EncodeVideo;

    return VLC_SUCCESS;

error:
    if( p_sys->e ) Engine_close( p_sys->e );
    free( p_sys->psz_ti_engine );
    free( p_sys );
    return VLC_EGENERIC;
}

/*****************************************************************************
 *
 *****************************************************************************/
void CloseVideoEncoder( vlc_object_t *p_this )
{
    encoder_t *p_enc = (encoder_t *)p_this;
    encoder_sys_t *p_sys = p_enc->p_sys;

    /* Delete video encoder */
    VIDENC_delete( p_sys->c );

    /* Delete codec engine handle */
    Engine_close( p_sys->e );

    /* Exit the codec engine */
    CERuntime_exit(); /* XXX: Should this be done only once (if we have encoder + decoder runing) */

    /* Free 'DaVinci compliant' buffers */
    davinci_FreeBuffer( &p_sys->in );
    davinci_FreeBuffer( &p_sys->out );

    free( p_sys->psz_ti_engine );
    free( p_sys );
}

/*****************************************************************************
 *
 *****************************************************************************/
static inline void davinci_CopyPictureToXDM( encoder_t *p_enc, XDM_BufDesc *p_buf, picture_t *p_pic )
{
    VLC_UNUSED( p_enc );

    /* Copy input picture */
    assert( p_pic->i_planes == p_buf->numBufs );
    assert( p_pic->i_planes == 1 );

    for( int i = 0; i < p_pic->i_planes; i++ )
    {
        plane_t *p = p_pic->p + i;
        memcpy( p_buf->bufs[i], p->p_pixels, p->i_pitch * p->i_visible_lines );
    }
}

static int davinci_InitBuffers( encoder_t *p_enc )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    int i_ret = VLC_SUCCESS;

    VIDENC_DynamicParams dparams;
    VIDENC_Status status;

    dparams.size = sizeof( dparams );
    memset( &status, 0, sizeof( status ) );
    status.size = sizeof( status );

    /* Configue the encoder */
    dparams.inputHeight = p_enc->fmt_in.video.i_height;
    dparams.inputWidth = p_enc->fmt_in.video.i_width;
    dparams.refFrameRate = p_enc->fmt_out.video.i_frame_rate_base ?
                            p_enc->fmt_out.video.i_frame_rate * 1000 /
                            p_enc->fmt_out.video.i_frame_rate_base :
                            p_enc->i_iframes * 1000; /* Frames per 1000 seconds  */
    dparams.targetFrameRate = dparams.refFrameRate; /* input fps = output fps */
    dparams.targetBitRate = p_enc->fmt_out.i_bitrate;
    dparams.intraFrameInterval = p_enc->i_iframes;
    dparams.generateHeader = XDM_ENCODE_AU; /* don't encode only the header */
    dparams.captureWidth = 0;
    dparams.forceIFrame = 0;

    if( VIDENC_control( p_sys->c, XDM_SETPARAMS, &dparams, &status )
        != VIDENC_EOK )
    {
        msg_Err( p_enc, "Failed to set encoder parameters: %dx%d @%d fps, "
            "%d kBps, I-frame interval %d",
            (int)dparams.inputWidth, (int)dparams.inputHeight,
            (int)dparams.refFrameRate / 1000,
            (int)dparams.targetBitRate >> 13 /* / (1024 * 8) */,
            (int)dparams.intraFrameInterval );
        return VLC_EGENERIC;
    }

    msg_Info( p_enc, "using %dx%d at %.3f fps (bitrate %d kBps, I-Frame interval %d)\n",
        (int)dparams.inputWidth, (int)dparams.inputHeight,
        ((float)dparams.targetFrameRate)/1000.,
        ((int)dparams.targetBitRate) >> 13 /* / (8*1024)*/,
        (int)dparams.intraFrameInterval );

    /* Configure buffers */
    if( VIDENC_control( p_sys->c, XDM_GETBUFINFO, &dparams, &status )
        != VIDENC_EOK )
    {
        msg_Err( p_enc, "Failed to get buffer info" );
        return VLC_EGENERIC;
    }

    /* Allocate input buffer(s) */
    if( (i_ret = davinci_AllocateBuffer( status.bufInfo.minNumInBufs,
                                  status.bufInfo.minInBufSize, &p_sys->in ))
        != VLC_SUCCESS )
    {
        msg_Err( p_enc, "Failed to allocate input buffers" );
        return i_ret;
    }

    /* Allocate output buffer */
    if( (i_ret = davinci_AllocateBuffer( status.bufInfo.minNumOutBufs,
                                  status.bufInfo.minOutBufSize, &p_sys->out ))
        != VLC_SUCCESS )
    {
        msg_Err( p_enc, "Failed to allocate input buffers" );
        return i_ret;
    }

    return i_ret;
}

static block_t *EncodeVideo( encoder_t *p_enc, picture_t *p_pic )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    block_t *p_block;
    VIDENC_InArgs in_args;
    VIDENC_OutArgs out_args;
    int i;

    if( p_sys->in.numBufs == 0 || p_sys->out.numBufs == 0 )
    {
        if( davinci_InitBuffers( p_enc ) != VLC_SUCCESS )
            return NULL;
    }

    davinci_CopyPictureToXDM( p_enc, &p_sys->in, p_pic );

    /* Configure input */
    in_args.size = sizeof( in_args );

    /* Configure output */
    memset( &out_args, 0, sizeof( out_args ) );
    out_args.size = sizeof( out_args );

    /* Encode the video */
    i = VIDENC_process( p_sys->c, &p_sys->in, &p_sys->out,
                        &in_args, &out_args );
    if( i != VIDENC_EOK )
    {
        msg_Err( p_enc, "Video encoding failed (%d): %s", i,
                 davinci_GetExtendedError( out_args.extendedError ) );
        return NULL;
    }

    /* Print some info */
    //msg_Dbg( p_enc, "Bytes generated: %d", (int)out_args.bytesGenerated );

    /* Put everything in the block */
    if( out_args.bytesGenerated <= 0 )
        return NULL;

    p_block = block_New( p_enc, out_args.bytesGenerated );
    if( !p_block )
        return NULL;

    memcpy( p_block->p_buffer, p_sys->out.bufs[0], out_args.bytesGenerated );

    switch( out_args.encodedFrameType )
    {
        case IVIDEO_I_FRAME:
        case IVIDEO_II_FRAME:
            p_block->i_flags |= BLOCK_FLAG_TYPE_I;
            break;
        case IVIDEO_P_FRAME:
        case IVIDEO_PP_FRAME:
            p_block->i_flags |= BLOCK_FLAG_TYPE_P;
            break;
        case IVIDEO_B_FRAME:
        case IVIDEO_BB_FRAME:
            p_block->i_flags |= BLOCK_FLAG_TYPE_B;
            break;
        default:
            msg_Warn( p_enc, "Unknown frame type" );
            p_block->i_flags |= BLOCK_FLAG_TYPE_PB; /* Wild guess */
            break;
    }

    p_block->i_dts = p_block->i_pts = p_pic->date;

    /* Shamelessly copied from ffmpeg/encoder.c */
    p_block->i_length = INT64_C(1000000) *
        p_enc->fmt_in.video.i_frame_rate_base /
            p_enc->fmt_in.video.i_frame_rate;

    return p_block;
}

