/*****************************************************************************
 * resizer.c: Resize pictures directly into framebuffer (DaVinci specific)
 *****************************************************************************
 * Copyright (C) 2008-2009 M2X BV
 *
 * Authors: Rafaël Carré <rcarre@m2x.nl>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#undef _FILE_OFFSET_BITS /* mmap() fails on 64 bits offsets */
#include "davinci.h" /* DAVINCI_HACK defined here */

#ifdef DAVINCI_HACK

#include "resizer.h"

#include <assert.h>

static void Resize_coeff( decoder_t *p_dec,
    unsigned int *i_width, unsigned int *i_height )
{
    unsigned int i_vidw = p_dec->fmt_out.video.i_width;
    unsigned int i_vidh = p_dec->fmt_out.video.i_height;
    unsigned int i_hratio = VOUT_ASPECT_FACTOR * *i_height / i_vidh;
    unsigned int i_wratio = VOUT_ASPECT_FACTOR * *i_width / i_vidw;

    msg_Dbg( p_dec, "Original Video %dx%d, Output destination %dx%d",
        i_vidw, i_vidh, *i_width, *i_height );

    /* Davinci Resizer can only do from x0.25 to x4 */
    if( i_hratio > VOUT_ASPECT_FACTOR * 4 )
        i_hratio = VOUT_ASPECT_FACTOR * 4;
    else if( i_hratio < VOUT_ASPECT_FACTOR / 4 )
        i_hratio = VOUT_ASPECT_FACTOR / 4;

    /* do that also for the width */
    if( i_wratio > VOUT_ASPECT_FACTOR * 4 )
        i_wratio = VOUT_ASPECT_FACTOR * 4;
    else if( i_wratio < VOUT_ASPECT_FACTOR / 4 )
        i_wratio = VOUT_ASPECT_FACTOR / 4;

    /* choose the minimum for both axes */
    i_hratio = i_wratio = i_hratio < i_wratio ? i_hratio : i_wratio;

    /* Now calculate hrsz and vrsz */
    int i_hrsz = 256 * VOUT_ASPECT_FACTOR / i_wratio;
    int i_vrsz = 256 * VOUT_ASPECT_FACTOR / i_hratio;

    /* Then recalculate final resolution */
    int i_htap = i_hrsz > 512 ? 7 : 4;
    int i_vtap = i_vrsz > 512 ? 7 : 4;

    *i_width = ((i_vidw - 7) * 256 - (i_htap ==  7 ? 32 : 16)) / i_hrsz + 1;
    *i_height = ((i_vidh - i_vtap) * 256 - (i_vtap ==  7 ? 32 : 16)) / i_vrsz + 1;

    /* width must be a factor of 16 : davinci resizer limitation when doing
     * vertical upscaling, 16 bytes per output line
     * the bpp is 2 so a factor of 8 should be good but not in practice
     */
    *i_width = ( *i_width + 15 ) & ~15;

    msg_Dbg( p_dec, "Destination fixed to %dx%d", *i_width, *i_height );
}


/* open() a framebuffer by its name (NOT device path) */
static int OpenFB( const char *psz_name )
{
    int i_fd = -1;
    int i_tries = 0; /* begin with /dev/fb0 */
    char *psz_device;
    struct fb_fix_screeninfo info;

    do {
        if( asprintf( &psz_device, "/dev/fb%d", i_tries++ ) == -1 )
            return -1;
        i_fd = open( psz_device, O_RDWR);
        free( psz_device );

        if( i_fd == -1 )
            return -1; /* check errno */

        if( ioctl( i_fd, FBIOGET_FSCREENINFO, &info ) == -1 )
        {
            close( i_fd );
            continue;
        }

        if( !strcmp( info.id, psz_name ) )
            return i_fd;

        close( i_fd );
    } while(1);
}

static float bicubic_core( float s )
{
    if( s < 0 ) s = -s;

    if( 0. <= s && s <= 1. )
        return 1.5 * s * s * s - 2.5 * s * s + 1.;
    else if( s <= 2. )
        return -.5 * s * s * s + 2.5 * s * s - 4. *s + 2.;
    else
        return 0.;
}


static void get_coeffs( short coefs[32], unsigned int i_in,
    unsigned int i_out )
{
    assert( i_out > 1 );
    assert( i_in > 7 );

    float f_deinterleave_coef[ 32 ];
    float f_windows[ 32+1 ];

    unsigned int i_rsz;

    if( i_in == i_out )
        i_rsz = 256;
    else
    {
        i_rsz = ((i_in - 7) * 256) / (i_out - 1);
        if( i_rsz <= 512 )
        {
            i_rsz = ((i_in - 4) * 256) / (i_out - 1);
            if( i_rsz > 512 ) i_rsz = 512;
        }
    }

    int i_nphases, i_ntaps, i_winlen, i_phase_offset, i, j;
    float f_fc;

    if( i_rsz > 512 )
    {
        i_nphases = 4;
        i_ntaps = 7;
        i_phase_offset = 8;
    }
    else
    {
        i_nphases = 8;
        i_ntaps = 4;
        i_phase_offset = 4;
    }
    i_winlen = i_nphases * i_ntaps;
    /* calculating the cut-off frequency, normalized by fs/2 */
    f_fc = ( i_rsz < 256) ? (1. / i_nphases) : 256. / (i_rsz * i_nphases);

    for( i = 0 ; i <= i_winlen ; i++ )
        f_windows[i] = .42 - .5 * cos( 2 * PI * i / i_winlen ) +
            .08 * cos( 4 * PI * i / i_winlen );

    i_nphases = 8;
    i_ntaps = 4;
    i_phase_offset = 4;
    /* calculating the bi-cubic coefficients */
    for( i = 0 ; i < i_nphases ; i++ )
    {
        float f_alpha = ((float)i) / i_nphases;

        f_deinterleave_coef[i_phase_offset*i] = bicubic_core( f_alpha + 1. );
        f_deinterleave_coef[i_phase_offset*i+1] = bicubic_core( f_alpha );
        f_deinterleave_coef[i_phase_offset*i+2] = bicubic_core( 1. - f_alpha );
        f_deinterleave_coef[i_phase_offset*i+3] = bicubic_core( 2. - f_alpha );
    }

    /* de-interleave into phases */
    for( i = 0 ; i < i_nphases ; i++ )
    { /* i=phase */
        float f_gain = 0;
        for( j = 0 ; j < i_ntaps ; j++ ) /* j=tap */
            f_gain += f_deinterleave_coef[i_phase_offset * i + j];
        for( j = 0 ; j < i_ntaps ; j ++)
            coefs[i_phase_offset * i + j] = (short)
                (f_deinterleave_coef[i_phase_offset*i+j] * 256 / f_gain + 0.5 );
        if( i_ntaps == 7 )
            coefs[i_phase_offset * i + 7] = 0.;
    }

    /* adjust the gain to make it exactly one */
    for( i = 0 ; i < i_nphases ; i++ )
    {
        int max = coefs[i_phase_offset*i];
        int max2 = 0;
        int index = 0;
        int index2 = 0;
        int total_gain = max;
        int delta;
        for( j = 1 ; j < i_ntaps ; j++ )
        {
            total_gain += coefs[i_phase_offset*i+j];
            if( abs( coefs[i_phase_offset*i+j] ) >= max )
            {
                max2 = max;
                index2 = index;
                index = j;
                max = abs( coefs[i_phase_offset*i+j] );
            }
        }
        delta = 256 - total_gain;

        if(max - max2 < 10)
        {
            coefs[i_phase_offset*i+index2] += (delta>>1);
            coefs[i_phase_offset*i+index] += (delta-(delta>>1));
        }
        else
            coefs[i_phase_offset*i+index] += delta;
    }

    if( i_rsz == 256 )
    { /* no resizing, all pass filtering */
        for( i = 0 ; i < i_nphases * i_ntaps ; i+= 4 )
        {
            coefs[i] = 256;
            coefs[i+1] = coefs[i+2] = coefs[i+3] = 0;
        }
    }
}

void Resize( decoder_t *p_dec, bool b_scale, davinci_resizer_t *rsz,
    davinci_fb_t *fb, XDM_BufDesc out )
{
    if( !rsz->i_yuyv )
    {
        /* use screen resolution for the buffer, we may be off by a few */
        rsz->i_yuyv = fb->var_info.xres * fb->var_info.yres_virtual;
        assert(rsz->i_yuyv);

        rsz_reqbufs_t reqbufs;
        rsz_buffer_t buffer;
        memset( &reqbufs, 0, sizeof( reqbufs ) );
        memset( &buffer, 0, sizeof( buffer ) );

        reqbufs.buf_type = RSZ_BUF_OUT;
        reqbufs.size = rsz->i_yuyv;
        reqbufs.count = 1;

        if( ioctl( rsz->i_fd, RSZ_REQBUF, &reqbufs ) )
        {
            msg_Err( p_dec, "RSZ_REQBUF failed (%m)" );
            return;
        }

        buffer.buf_type = RSZ_BUF_OUT;
        buffer.index = 0;

        if( ioctl( rsz->i_fd, RSZ_QUERYBUF, &buffer ) )
        {
            msg_Err( p_dec, "RSZ_QUERYBUF failed (%m)" );
            return;
        }

        rsz->offset = buffer.offset;

        rsz->p_yuyv = mmap( NULL, reqbufs.size, PROT_READ|PROT_WRITE,
                 MAP_SHARED, rsz->i_fd, buffer.offset );
        if( rsz->p_yuyv == MAP_FAILED )
        {
            rsz->i_yuyv = 0;
            msg_Err( p_dec, "mmap resizer to buffer failed (%m)" );
            return;
        }
    }

    /* Sets resizer parameters when video starts or when resolution changes */
    if( rsz->i_width != p_dec->fmt_out.video.i_width ||
        rsz->i_height != p_dec->fmt_out.video.i_height )
    {
        rsz_params_t rsz_params;
        memset( &rsz_params, 0, sizeof( rsz_params ) );
        rsz_params.in_hsize = rsz->i_width = p_dec->fmt_out.video.i_width;
        rsz_params.in_vsize = rsz->i_height = p_dec->fmt_out.video.i_height;
        rsz_params.in_pitch =
            (p_dec->fmt_out.video.i_width * BPP / 8 + 31) & ~31;
        rsz_params.inptyp = RSZ_INTYPE_YCBCR422_16BIT;
        rsz_params.pix_fmt = RSZ_PIX_FMT_YUYV;

        /* Sets the desired maximum destination resolution */
        if( b_scale )
        {
            rsz_params.out_hsize = fb->var_info.xres;
            rsz_params.out_vsize = fb->var_info.yres;
        }
        else
        {
            rsz_params.out_hsize = rsz_params.in_hsize;
            rsz_params.out_vsize = rsz_params.in_vsize;
        }

        /* And then modify it to keep the same aspect ratio, and make sure
         * that the hardware is able to handle it */
        Resize_coeff( p_dec, &rsz_params.out_hsize, &rsz_params.out_vsize );

        /* RSZ_PIX_FMT_YUYV is 2 bytes per pixel */
        rsz_params.out_pitch = ( rsz_params.out_hsize * 2 + 15 ) & ~15;

        rsz->b_direct =
            (unsigned)rsz_params.out_hsize == (unsigned)fb->var_info.xres;

        rsz->i_out_width = rsz_params.out_hsize;
        rsz->i_out_height = rsz_params.out_vsize;

        short hcoefs[32], vcoefs[32];

        get_coeffs( hcoefs , rsz_params.in_hsize, rsz_params.out_hsize );
        get_coeffs( vcoefs , rsz_params.in_vsize, rsz_params.out_vsize );

        int i;
        for( i = 0; i < 32; i++ )
        {
            rsz_params.hfilt_coeffs[i] = hcoefs[i];
            rsz_params.vfilt_coeffs[i] = vcoefs[i];
        }

        rsz_params.yenh_params.type = RSZ_YENH_DISABLE;

        if( ioctl( rsz->i_fd, RSZ_S_PARAM, &rsz_params ) )
        {
            msg_Err( p_dec, "Failed setting resizer parameters: "
                "%dx%d (pitch %d) -> %dx%d (pitch %d) : %m",
                rsz_params.in_hsize, rsz_params.in_vsize, rsz_params.in_pitch,
                rsz_params.out_hsize, rsz_params.out_vsize, rsz_params.out_pitch
            );
            rsz->i_width = rsz->i_height = 0;

            return;
        }
    }

    rsz_resize_t rszh;
    rszh.in_buf.index     = -1;
    rszh.in_buf.buf_type  = RSZ_BUF_IN;
    rszh.in_buf.size      =
        p_dec->fmt_out.video.i_height *
        ((p_dec->fmt_out.video.i_width * BPP / 8 + 31) & ~31);
    rszh.in_buf.offset    = Memory_getBufferPhysicalAddress(
        out.bufs[0], rszh.in_buf.size, NULL );
    assert( rszh.in_buf.offset );

    rszh.out_buf.index    = -1;
    rszh.out_buf.buf_type = RSZ_BUF_OUT;

    if( b_scale && rsz->b_direct /* output video res is screen res */ )
    {
        /* FIXME: use NUM_BUFFERS if needed */
        rszh.out_buf.offset   = fb->p_physbufs[0];
        rszh.out_buf.size     = fb->var_info.xres * fb->var_info.yres_virtual;
    }
    else
    {
        rszh.out_buf.offset  =   rsz->offset;
        rszh.out_buf.size    =   rsz->i_yuyv;
    }

    do
    {
        if( ioctl( rsz->i_fd, RSZ_RESIZE, &rszh ) == -1 )
        {
            if( errno == EAGAIN )
                continue;

            msg_Err( p_dec, "Resizing failed (%m)" );
            return;
        }
    } while(0);
}

int DavinciInit( decoder_t *p_dec, davinci_fb_t *fb, davinci_resizer_t *rsz )
{
    fb->i_fd = rsz->i_fd = -1;

    /* Open framebuffer */
    if( ( fb->i_fd = OpenFB( "dm_vid0_fb" ) ) == -1 )
    {
        msg_Err( p_dec, "Failed to open framebuffer (%m)" );
        goto error;
    }

    if( ioctl( fb->i_fd, FBIOGET_VSCREENINFO, &fb->var_info ) == -1 )
    {
        msg_Err( p_dec, "Failed to get framebuffer info (%m)" );
        goto error;
    }

    fb->var_info.yres_virtual = fb->var_info.yres * NUM_BUFFERS;
    fb->var_info.bits_per_pixel = BPP;

    if( ioctl( fb->i_fd, FBIOPUT_VSCREENINFO, &fb->var_info ) == -1 )
    {
        msg_Err( p_dec, "Can not set fb info (%m)" );
        goto error;
    }

    if( fb->var_info.bits_per_pixel != BPP )
    {
        msg_Err( p_dec, "Can not use requested size (%dx%d at %d bpp)",
            fb->var_info.xres, fb->var_info.yres,
            fb->var_info.bits_per_pixel );
        goto error;
    }

    if( ioctl( fb->i_fd, FBIOGET_FSCREENINFO, &fb->fix_info ) == -1 )
    {
        msg_Err( p_dec, "Failed to get framebuffer fixed info (%m)" );
        goto error;
    }

    /* Yes, 32bits * 32bits fills 64 bits.
 *      * Oh and if you find a screen with 16 billions of billions of pixels ...
 *           * Keep it for yourself. */
    uint64_t i_display_sz = fb->fix_info.line_length * fb->var_info.yres;
    uint64_t i_display_total_sz =
        fb->fix_info.line_length * fb->var_info.yres_virtual;

    fb->p_map = mmap( NULL, i_display_total_sz,
        PROT_READ | PROT_WRITE, MAP_SHARED, fb->i_fd, 0 );

    if( fb->p_map == MAP_FAILED )
    {
        fb->p_map = NULL; /* set to NULL for error handling */
        msg_Err( p_dec, "Memory mapping of framebuffer failed (%m)" );
        goto error;
    }
    /* Clears fb */
    unsigned int i;
    for( i = 0 ; i < i_display_total_sz / 4 ; i++ )
        ((uint32_t*)fb->p_map)[i] = 0x10801080;

    for( i = 0 ; i < NUM_BUFFERS ; i++ )
        if( ( fb->p_physbufs[i] = Memory_getBufferPhysicalAddress(
            fb->p_map + i * i_display_sz, i_display_sz, NULL ) ) == 0 )
        {
            msg_Err( p_dec, "Unable to get physical address of buffer "
                "%d/%d (%"PRId64" bytes)", i, NUM_BUFFERS, i_display_sz );
            goto error;
        }

    /* Open and setup resizer */
    if( ( rsz->i_fd = open( "/dev/davinci_resizer", O_RDWR ) ) == -1 )
    {
        msg_Err( p_dec, "Failed to open resizer (%m)" );
        goto error;
    }

    rsz->p_yuyv = NULL;
    rsz->i_yuyv = 0;

    rsz->i_height = rsz->i_width = 0;

    return VLC_SUCCESS;

error:
    DavinciClose( fb, rsz );
    return VLC_EGENERIC;
}

void DavinciClose( davinci_fb_t *fb, davinci_resizer_t *rsz )
{
    if( fb->p_map )
    {
        /* Clears fb */
        unsigned int i;
        uint64_t i_display_size =
            fb->fix_info.line_length * fb->var_info.yres_virtual / 4;

        for( i = 0 ; i < i_display_size ; i++ )
            ((uint32_t*)fb->p_map)[i] = 0x10801080;

        munmap( fb->p_map, i_display_size );
    }

    if( fb->i_fd != -1 ) close( fb->i_fd );
    if( rsz->i_fd != -1 ) close( rsz->i_fd );

    rsz->i_fd = fb->i_fd = -1;
    fb->p_map = NULL;
}

#endif /* DAVINCI_HACK */
