/*****************************************************************************
 * video.c: video decoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008-2009 M2X BV
 * $Id: df57dc519fa74927d4c883116a8be975662f40b2 $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *          Rafaël Carré <rcarre@m2x.nl>
 *          Jean-Paul Saman <jean-paul.saman at m2x dot nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_codec.h>
#include <vlc_vout.h>
#include <vlc_codecs.h>

#include "davinci.h"

#include <ti/sdo/ce/video/viddec.h>

#include <assert.h>

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static picture_t *DecodeVideoBlock( decoder_t *, block_t ** );

/*****************************************************************************
 * Video decoder
 *****************************************************************************/
struct decoder_sys_t
{
    char *psz_ti_engine;
    Engine_Handle e;
    VIDDEC_Handle d;

    XDM_BufDesc in;
    XDM_BufDesc out;
};

/*****************************************************************************
 * OpenDecoder: probe the decoder and return score
 *****************************************************************************/
int OpenVideoDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t*)p_this;
    decoder_sys_t *p_sys;
    Engine_Error err;
    int i_cat;
    char *psz_chroma = NULL;
    const char *psz_codec = NULL;
    const char *psz_namecodec = NULL;
    vlc_fourcc_t i_chroma = VLC_FOURCC('U','Y','V','Y');
    VIDDEC_Params params;

    if( !GetDavinciDecoder( p_dec->fmt_in.i_codec, &i_cat,
                            &psz_codec, &psz_namecodec ) )
        return VLC_EGENERIC;

    msg_Dbg( p_dec, "found davinci decoder %s for %s (codec=%4.4s, cat=%s)",
                psz_codec, (char*)&p_dec->fmt_out.i_codec, psz_namecodec,
                (i_cat == VIDEO_ES) ? "video" :
                    (i_cat == AUDIO_ES) ? "audio" : "other" );

    /* Allocate our private structure */
    p_dec->p_sys = (decoder_sys_t *)calloc( 1, sizeof( decoder_sys_t ) );
    if( !p_dec->p_sys )
        return VLC_ENOMEM;

    p_sys = p_dec->p_sys;

    p_sys->psz_ti_engine = var_CreateGetString( p_dec, DEC_CFG_PREFIX "engine" );
    if( !p_sys->psz_ti_engine )
    {
        free( p_sys );
        return VLC_ENOMEM;
    }

    /* Initialize the codec engine */
    CERuntime_init();

    /* Create an engine handle */
    p_sys->e = Engine_open( p_sys->psz_ti_engine, NULL /*&Engine_ATTRS*/, &err );
    if( err != Engine_EOK )
    {
        msg_Err( p_dec, "Error while opening engine `%s': %s",
                 p_sys->psz_ti_engine, ppsz_engine_error[err] );
        goto error;
    }

    davinci_PrintAvailableAlgorithms( p_this, p_sys->psz_ti_engine );

    /* Get user supplied chroma setting */
    psz_chroma = var_CreateGetString( p_dec, DEC_CFG_PREFIX "fourcc" );
    if( psz_chroma && strlen( psz_chroma ) == 4 )
    {
        i_chroma = VLC_FOURCC( psz_chroma[0], psz_chroma[1],
                               psz_chroma[2], psz_chroma[3] );
        msg_Dbg( p_dec, "Forcing output chroma to `%s'", psz_chroma );
    }
    free( psz_chroma );

    /* Create video decoder */
    params.size = sizeof( VIDDEC_Params );

    /* TODO change the params when HD codecs are available */
    params.maxHeight = p_dec->fmt_in.video.i_height ?: 576 /*720*/;
    params.maxWidth = p_dec->fmt_in.video.i_width ?: 720 /*1280*/;
    /* The application should set the parameters to 0 to use the algorithm's default values. */
    params.maxFrameRate = 0;
    params.maxBitRate = 0;

    msg_Dbg( p_dec, "Creating decoder for %dx%d codec %4.4s in chroma %4.4s",
        (int)params.maxWidth, (int)params.maxHeight,
        (const char *)&p_dec->fmt_in.i_codec,
        (char*)&i_chroma
        );

    params.dataEndianness = XDM_BYTE;
    params.forceChromaFormat = VlcChromaToXdm( i_chroma );
    if( params.forceChromaFormat == XDM_CHROMA_NA )
    {
        msg_Err( p_dec, "Unsupported output chroma (%4.4s).",
                 (const char *)&i_chroma );
        goto error;
    }

    p_sys->d = VIDDEC_create( p_sys->e, (String)psz_codec, &params );
    if( !p_sys->d )
    {
        msg_Err( p_dec, "Failed to create video decoder (%s) for %s",
                 psz_codec, psz_namecodec );
        goto error;
    }

    /* Set output properties */
    p_dec->fmt_out.i_cat = VIDEO_ES;
    p_dec->fmt_out.i_codec =
        p_dec->fmt_out.video.i_chroma = i_chroma;
    p_dec->fmt_out.video.i_width = p_dec->fmt_in.video.i_width;
    p_dec->fmt_out.video.i_height = p_dec->fmt_in.video.i_height;
    p_dec->fmt_out.video.i_aspect = p_dec->fmt_in.video.i_aspect;

    /* Set callbacks */
    p_dec->pf_decode_video = DecodeVideoBlock;

    /* DaVinci decoder needs complete frames */
    p_dec->b_need_packetized = true;
    return VLC_SUCCESS;

error:
    if( p_sys->e ) Engine_close( p_sys->e );
    free( p_sys->psz_ti_engine );
    free( p_sys );
    return VLC_EGENERIC;
}

/*****************************************************************************
 * CloseDecoder: png decoder destruction
 *****************************************************************************/
void CloseVideoDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t *)p_this;
    decoder_sys_t *p_sys = p_dec->p_sys;

    /* Close our codec handle */
    VIDDEC_delete( p_sys->d );

    /* Close our engine handle */
    Engine_close( p_sys->e );

    /* Exit the codec engine */
    CERuntime_exit();

    /* Free 'DaVinci compliant' buffers */
    davinci_FreeBuffer( &p_sys->in );
    davinci_FreeBuffer( &p_sys->out );

    free( p_sys->psz_ti_engine );
    free( p_sys );
}

static inline void davinci_CopyXDMPicture( decoder_t *p_dec, picture_t *p_pic, XDM_BufDesc *p_buf )
{
    /* Copy stuff to picture here */
    assert( p_pic->i_planes == p_buf->numBufs );

    for( int i = 0; i < p_pic->i_planes; i++ )
    {
        plane_t *p = p_pic->p + i;
        memcpy( p->p_pixels, p_buf->bufs[i], p->i_pitch * p->i_visible_lines );
    }
}

/****************************************************************************
 * DecodeBlock: the whole thing
 ****************************************************************************/
static picture_t *DecodeVideoBlock( decoder_t *p_dec, block_t **pp_block )
{
    decoder_sys_t *p_sys = p_dec->p_sys;
    block_t *p_block;
    picture_t *p_pic = NULL;
    VIDDEC_InArgs in_args;
    VIDDEC_OutArgs out_args;
    VIDDEC_DynamicParams dparams;
    VIDDEC_Status status;

    if( !pp_block || !*pp_block ) return NULL;
    p_block = *pp_block;

    if( p_block->i_flags & (BLOCK_FLAG_DISCONTINUITY|BLOCK_FLAG_CORRUPTED) )
    {
        block_Release( p_block );
        return NULL;
    }

    memset( &in_args, 0, sizeof( in_args ) );
    memset( &out_args, 0, sizeof( out_args ) );

    /* Configure video decoder */
    dparams.size = sizeof( VIDDEC_DynamicParams );
    memset( &status, 0, sizeof( VIDDEC_Status ) );
    status.size = sizeof( VIDDEC_Status );

    if( p_sys->in.numBufs == 0 || p_sys->out.numBufs == 0 )
    {
        if( VIDDEC_control( p_sys->d, XDM_GETBUFINFO, &dparams, &status )
            != VIDDEC_EOK )
        {
            msg_Err( p_dec, "Failed to get buffer info" );
            goto error;
        }
    }
    if( p_sys->in.numBufs == 0 )
    {
        /* Allocate input buffer */
        if( davinci_AllocateBuffer( status.bufInfo.minNumInBufs,
                                    status.bufInfo.minInBufSize, &p_sys->in )
            != VLC_SUCCESS )
        {
            msg_Err( p_dec, "Failed to allocate input buffers" );
            goto error;
        }
    }
    if( p_sys->out.numBufs == 0 )
    {
        /* Allocate output buffer(s) */
        if( davinci_AllocateBuffer( status.bufInfo.minNumOutBufs,
                                    status.bufInfo.minOutBufSize, &p_sys->out )
            != VLC_SUCCESS )
        {
            davinci_FreeBuffer( &p_sys->in );
            msg_Err( p_dec, "Failed to allocate output buffers" );
            goto error;
        }
    }

    if( VIDDEC_control( p_sys->d, XDM_GETSTATUS, &dparams, &status )
        != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to get decoder status" );
        goto error;
    }

    /* Setup input arguments */
    in_args.size = sizeof( in_args );
    in_args.numBytes = __MIN( p_dec->fmt_in.i_extra + p_block->i_buffer,
                              (size_t)p_sys->in.bufSizes[0] );
    in_args.inputID = 1;

    /* Setup input buffer */
    assert( p_block->i_buffer <= (size_t)p_sys->in.bufSizes[0] );

    if( p_dec->fmt_in.i_extra > 0 )
    {
        memcpy( p_sys->in.bufs[0], p_dec->fmt_in.p_extra, p_dec->fmt_in.i_extra );
    }
    memcpy( p_sys->in.bufs[0] + p_dec->fmt_in.i_extra, p_block->p_buffer,
            in_args.numBytes - p_dec->fmt_in.i_extra );

#ifdef DEBUG_DAVINCI
    msg_Dbg( p_dec, "Frame is %d bytes (%d extra), %dx%d, type %s",
            (int)in_args.numBytes, p_dec->fmt_in.i_extra, (int)status.outputWidth,
            (int)status.outputHeight,
              status.contentType == IVIDEO_PROGRESSIVE ? "progressive" :
              status.contentType == IVIDEO_INTERLACED ? "interlaced" :
              "unknown" );
#endif

    /* Setup output arguemnts */
    out_args.size = sizeof( out_args );

    /* Decode the video */
    if( VIDDEC_process( p_sys->d, &p_sys->in, &p_sys->out, &in_args, &out_args ) != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Video decoding failed: %s",
            davinci_GetExtendedError(out_args.extendedError ) );
        goto error;
    }

    if( VIDDEC_control( p_sys->d, XDM_GETSTATUS, &dparams, &status ) != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to get decoder status" );
        goto error;
    }

    if( in_args.numBytes > out_args.bytesConsumed )
        msg_Err( p_dec, "%d bytes were not consumed", (int)(out_args.bytesConsumed - in_args.numBytes) );

    p_block->p_buffer += out_args.bytesConsumed - p_dec->fmt_in.i_extra;
    p_block->i_buffer -= out_args.bytesConsumed - p_dec->fmt_in.i_extra;

    p_dec->fmt_out.video.i_width = status.outputWidth;
    p_dec->fmt_out.video.i_height = status.outputHeight;
    p_dec->fmt_out.video.i_aspect = VOUT_ASPECT_FACTOR * status.outputWidth
                                    / status.outputHeight; /* FIXME */

    /* Get a new picture */
    p_pic = decoder_NewPicture( p_dec );
    if( !p_pic )
    {
        msg_Err( p_dec, "Could not get picture" );
        goto error;
    }
    p_pic->b_progressive = (status.contentType == IVIDEO_INTERLACED) ?
                            false : true;

#ifdef DEBUG_DAVINCI
    switch( out_args.decodedFrameType )
    {
        case IVIDEO_NA_FRAME:
            msg_Dbg( p_dec, "Unknown frame type" );
            break;
        case IVIDEO_I_FRAME:
            msg_Dbg( p_dec, "I frame" );
            break;
        case IVIDEO_P_FRAME:
            msg_Dbg( p_dec, "P frame" );
            break;
        case IVIDEO_B_FRAME:
            msg_Dbg( p_dec, "B frame" );
            break;
        case IVIDEO_II_FRAME:
            msg_Dbg( p_dec, "Interlaced I frame" );
            break;
        case IVIDEO_PP_FRAME:
            msg_Dbg( p_dec, "Interlaced P frame" );
            break;
        case IVIDEO_BB_FRAME:
            msg_Dbg( p_dec, "Interlaced B frame" );
            break;
        default:
            msg_Dbg( p_dec, "Other frame type: %d", (int)out_args.decodedFrameType );
            break;
    }
#endif

#if 1
    /* Copy stuff to picture here */
    assert( p_pic->i_planes == p_sys->out.numBufs );

    for( int i = 0; i < p_pic->i_planes; i++ )
    {
        plane_t *p = p_pic->p + i;
        memcpy( p->p_pixels, p_sys->out.bufs[i], p->i_pitch * p->i_visible_lines );
    }
#else
    davinci_CopyXDMPicture( p_dec, p_pic, p_sys->out );
#endif

    /* */
    p_pic->date = p_block->i_pts > 0 ? p_block->i_pts : p_block->i_dts;

    if( p_block->i_buffer == 0 )
    {
        block_Release( p_block );
        *pp_block = NULL;
    }

    return p_pic;

error:
    if( p_pic )
        decoder_DeletePicture( p_dec, p_pic );

    block_Release( p_block );
    return NULL;
}
